package Fairness.Graphing;

import edu.rit.numeric.plot.Dots;
import edu.rit.numeric.plot.Grid;
import edu.rit.numeric.plot.Label;
import edu.rit.numeric.plot.LinearAxis;
import edu.rit.numeric.plot.Strokes;
import edu.rit.numeric.plot.Ticks;
import edu.rit.numeric.plot.XYPlot;
import edu.rit.swing.Displayable;
import edu.rit.swing.DisplayableIO;

import java.awt.*;
import java.io.File;
import java.io.IOException;
import java.text.DecimalFormat;
import java.util.List;

/**
 * Graph represents a graphical plot of 1 to 4 data series. A graph is created and customized
 * entirely through its constructor, then the image for the graph can be written to a file.
 *
 * @author Jason Winnebeck
 */
public class Graph {
    private static final Color green = new Color( 0, 128, 0 );

    private static final PlotSeriesInfo[] infos = new PlotSeriesInfo[] {
	    new PlotSeriesInfo( Dots.circle( Color.blue ),   Strokes.solid( 1.0 ), Color.blue  ),
	    new PlotSeriesInfo( Dots.diamond( Color.black ), Strokes.solid( 1.0 ), Color.black ),
	    new PlotSeriesInfo( Dots.square( Color.red ),    Strokes.solid( 1.0 ), Color.red   ),
	    new PlotSeriesInfo( Dots.circle( green      ),   Strokes.solid( 1.0 ), green       )
    };

    private final Displayable plot;

    /**
     * Constructs a new customized Graph. The named of the series is used to build a legend at the
     * bottom of the graph. Currently only up to 4 series are supported.
     *
     * @param title       non-null String with graph title
     * @param leftLabel   non-null String with y axis label
     * @param bottomLabel non-null String with x axis label
     * @param data        non-null List of 1 to 4 NamedXYSeries
     * @param bounds      non-null GraphBounds
     */
    public Graph( String title, String leftLabel, String bottomLabel,
                  List<? extends NamedXYSeries> data, GraphBounds bounds ) {
        final float titleFontSize = 18.0f;
        final float unitsFontSize = 16.0f;
        final Font titleFont =
            Label.DEFAULT_FONT.deriveFont( titleFontSize ).deriveFont( Font.BOLD );
        final Font unitsFont =
            Label.DEFAULT_FONT.deriveFont( unitsFontSize ).deriveFont( Font.BOLD );

        final double xLeft = bounds.getXLeft();
        final double xRight = bounds.getXRight();
        final double xMid = ( xLeft + xRight ) / 2.0;

        final double yBottom = bounds.getYBottom();
        final double yTop = bounds.getYTop();
        final double yMid = ( yBottom + yTop ) / 2.0;

        DecimalFormat theLabelFormat = new DecimalFormat( "0.000" );
        theLabelFormat.setMinimumFractionDigits( 0 );
        Ticks ticks = new Ticks( theLabelFormat );
        LinearAxis xAxis = new LinearAxis( xLeft, xRight, 10, 4, xLeft, 600, ticks, null );
        LinearAxis yAxis = new LinearAxis( yBottom, yTop, 10, 4, yBottom, 600, ticks, null );

        Grid grid = new Grid( Strokes.solid( 0.5 ), Color.BLACK,
                              Strokes.dashed( 0.5, 6.0, 3.0 ), Color.GRAY );

        XYPlot graph = new XYPlot( xAxis, yAxis, grid, 64.0, 64.0, 64.0, 84.0 );
        for ( int i = 0; i < data.size(); ++i ) {
            graph.addPlotSeries( infos[i].makePlotSeries( data.get( i ) ) );
        }

        Label label = new Label( title, xMid, yTop, Label.ABOVE, 4.0, titleFont );
        graph.addLabel( label );

        label = new Label( bottomLabel, xMid, yBottom, Label.BELOW, unitsFontSize * 2, unitsFont );
        graph.addLabel( label );

        label = new Label( leftLabel, xLeft, ( yBottom * 2 + yTop ) / 3,
                           Label.ROTATE_LEFT, unitsFontSize * 2, unitsFont );
        graph.addLabel( label );

        //Set up the Legend now
        Legend legend = new Legend( graph );
        for ( int i = 0; i < data.size(); ++i ) {
            legend.addLegendItem( infos[i].makeLegendItem( data.get( i ) ) );
        }

        plot = legend;
    }

    /**
     * Returns the Displayable graph object so that the Graph can be used in alternate rendering
     * contexts.
     */
    public Displayable getDisplayable() {
        return plot;
    }

    /**
     * Writes to the given file a color PNG image of the graph. The generated image is always
     * 728 by 748 pixels.
     *
     * @param outputFile non-null location of file to write
     *
     * @throws IOException if an error occurs trying to write the file
     */
    public void writeColorPng( File outputFile ) throws IOException {
        DisplayableIO.writeColorPNGFile( plot, outputFile );
    }

    /**
     * Writes to the given file a postscript image of the graph. The generated image is always
     * 728 by 748 pixels.
     *
     * @param outputFile non-null location of file to write
     *
     * @throws IOException if an error occurs trying to write the file
     */
    public void writePostscript( File outputFile ) throws IOException {
        DisplayableIO.writePostScriptFile( plot, outputFile );
    }
}
