package Fairness.Graphing;

import edu.rit.swing.Displayable;
import org.gillius.Util.CheckUtils;

import java.awt.*;
import java.awt.geom.Rectangle2D;
import java.awt.geom.AffineTransform;
import java.util.ArrayList;
import java.util.List;

/**
 * Legend provides a Displayable legend that is drawn on the bottom of another Displayable (such as
 * a {@link Graph}). Currently this class is used only internally in the Graph to draw its legend.
 *
 * @author Jason Winnebeck
 */
public class Legend implements Displayable {
    private final Displayable host;

    private final List<Displayable> items = new ArrayList<Displayable>();

    /**
     * Creates a new Legend that draws over the given host.
     */
    public Legend( Displayable host ) {
        CheckUtils.checkNull( host, "host" );

        this.host = host;
    }

    /**
     * Adds an item to be displayed in the Legend -- normally a {@link LegendItem}.
     * 
     * @param d non-null Displayable
     */
    public void addLegendItem( Displayable d ) {
        CheckUtils.checkNull( d, "d" );

        items.add( d );
    }

    /**
     * Returns this displayable object's bounding box. This is the smallest rectangle that encloses all
     * of this displayable object.
     */
    public Rectangle2D getBoundingBox() {
        return host.getBoundingBox();
    }

    /**
     * Returns this displayable object's background paint.
     */
    public Paint getBackgroundPaint() {
        return host.getBackgroundPaint();
    }

    /**
     * Draw this drawable object in the given graphics context. Upon return from this method, the given
     * graphics context's state (color, font, transform, clip, and so on) is the same as it was upon
     * entry to this method.
     *
     * @param g2d 2-D graphics context.
     */
    public void draw( Graphics2D g2d ) {
        CheckUtils.checkNull( g2d, "g2d" );

        host.draw( g2d );

        double width = 0;
        double height = 0;
        for ( Displayable item : items ) {
            width += item.getBoundingBox().getWidth();
            height = Math.max( height, item.getBoundingBox().getHeight() );
        }

        AffineTransform oldTransform = g2d.getTransform();

        Rectangle2D hostBox = host.getBoundingBox();
        double x = hostBox.getWidth() / 2 - width / 2;
        double y = hostBox.getHeight() - height;
        for ( Displayable item : items ) {
            g2d.translate( x, y );
            item.draw( g2d );
            g2d.setTransform( oldTransform );
            x += item.getBoundingBox().getWidth();
        }

        g2d.setTransform( oldTransform );
    }
}
