package Fairness;

import org.gillius.Util.CheckUtils;

import java.util.List;
import java.util.ArrayList;

/**
 * NodeMonitor contains methods to measure statistical information in the simulation.
 *
 * @author Jason Winnebeck
 */
public class NodeMonitor {
    private final String name;

    private Mark currentMark = new Mark();

    private final List<Mark> marks = new ArrayList<Mark>();

    /**
     * Creates an unnamed NodeMonitor. Used only for testing.
     */
    public NodeMonitor() {
        this( "Unnamed" );
    }

    /**
     * Creates a NodeMonitor with a given name.
     *
     * @param name non-null String
     */
    public NodeMonitor( String name ) {
        CheckUtils.checkNull( name, "name" );

        this.name = name;
    }

    /**
     * Returns this monitor's name
     */
    public String getName() {
        return name;
    }

    /**
     * Reports to the monitor that a packet was received.
     */
    public void reportReceived() {
        currentMark.received++;
    }

    /**
     * Reports to the monitor that a packet was lost.
     */
    public void reportLost() {
        currentMark.lost++;
    }

    /**
     * Saves currently accumulated statistics in a Mark and resets current statistics.
     */
    public void mark() {
        marks.add( currentMark );
        currentMark = new Mark();
    }

    /**
     * Returns a list of all marks created by {@link #mark()}.
     *
     * @return non-null List of 0 or more {@link Mark}.
     */
    public List<Mark> getMarks() {
        return marks;
    }

    /**
     * A Mark represents collected statistics from a period of a NodeMonitor's collection.
     */
    public static class Mark {
        private int received = 0;
        private int lost = 0;

        /**
         * Returns number of received packets.
         */
        public int getReceived() {
            return received;
        }

        /**
         * Returns percentage of received packets.
         */
        public double getReceivedPercent() {
            return (double)received / getTotal();
        }

        /**
         * Returns number of lost or unsent packets.
         */
        public int getLost() {
            return lost;
        }

        /**
         * Returns percentage of lost or unsent packets.
         */
        public double getLostPercent() {
            return (double)lost / getTotal();
        }

        /**
         * Returns total number of packets: received + lost.
         */
        public int getTotal() {
            return received + lost;
        }
    }
}
