package Fairness.Nodes;

import Fairness.NodeMonitor;
import Fairness.NodeTransferEvent;
import Fairness.Nodes.NodeBehaviors.NodeBehavior;
import Fairness.Nodes.TreeBehaviors.TreeBehavior;
import Fairness.Simulation;
import Fairness.SimulationParameters;
import org.gillius.Util.CheckUtils;

import java.util.ArrayList;
import java.util.HashMap;
import java.util.List;
import java.util.Map;
import java.util.Set;

/**
 * The Node class is the center of the program design. It contains all of the information for its
 * participation in the trees, for which there is one tree per stripe. A Node is typically in
 * multiple trees at the same time, depending on its amount of inbound bandwidth. The NodeBehavior
 * class is utilized to implement a freeloader or a freeloader detector of our two different types
 * (ancestor rating or debt maintenance). When packets are received there is a report made to the
 * NodeMonitor.
 *
 * @author Jason Winnebeck
 */
public class Node {
    private final int nodeId;
    private final Simulation sim;
    private final NodeMonitor monitor;

    private final NodeBehavior nodeBehavior;
    private final TreeBehavior treeBehavior;

    private int stripeId;

    private final int outboundBW;
    private final int inboundBW;

    private double lossRate = 0.0;

    private Map<Integer, TreeInfo> trees = new HashMap<Integer, TreeInfo>();

    /**
     * Constructs a new Node. After construction there still may be more setup to perform using the
     * {@link #setLossRate(double)} method.
     *
     * @param sim          non-null Simulation this Node will be part of
     * @param nodeId       Node ID as assigned from the TreeManager
     * @param outboundBW   outbound bandwidth
     * @param inboundBW    inbound bandwidth
     * @param monitor      non-null NodeMonitor
     * @param nodeBehavior non-null NodeBehavior
     * @param treeBehavior non-null TreeBehavior
     */
    public Node( Simulation sim, int nodeId, int outboundBW, int inboundBW, NodeMonitor monitor,
                 NodeBehavior nodeBehavior, TreeBehavior treeBehavior ) {
        CheckUtils.checkNull( sim, "sim" );
        CheckUtils.checkNull( monitor, "monitor" );
        CheckUtils.checkNull( nodeBehavior, "nodeBehavior" );
        CheckUtils.checkNull( treeBehavior, "treeBehavior" );

        this.sim = sim;
        this.nodeId = nodeId;
        this.inboundBW = inboundBW;
        this.outboundBW = outboundBW;
        this.monitor = monitor;

        this.nodeBehavior = nodeBehavior;
        this.treeBehavior = treeBehavior;
    }

    /**
     * Sets the percentage of packets received or sent that are lost due to network errors.
     *
     * @param lossRate a percentage (0 to 1) of packet loss.
     */
    public void setLossRate( double lossRate ) {
        if ( lossRate < 0.0 || lossRate > 1.0 )
          throw new IllegalArgumentException( "Parameter lossRate to Node.setLossRate must be within the range 0.0-1.0, inclusive; its value is " + lossRate + '.' );
        this.lossRate = lossRate;
    }

    /**
     * Processes a NodePacket that is being received by this Node during a Simulation. The
     * NodePacket belongs to a stripe and its source should equal the parent for that stripe.
     * <p/>
     * Node reports to the NodeMonitor whether the given packet is received or lost.
     * <p/>
     * A Node will route its packet to all of its children, with the sent and lost flags set
     * appropriately. The NodeBehavior for this Node is queried to see if the packet should be sent
     * to the node.
     *
     * @param sim    non-null Simulation
     * @param packet non-null NodePacket
     */
    public void processPacket( Simulation sim, NodePacket packet ) {
        CheckUtils.checkNull( sim, "sim" );
        CheckUtils.checkNull( packet, "packet" );

        if ( packet.getSource() != getParent( packet.getStripeId() ) )
            throw new AssertionError( "Invalid source for packet" );

        boolean lost = packet.isLost();
        boolean sent = packet.isSent();
        if ( !lost && sent && lossRate > 0 )
            lost = sim.getRng().nextDouble() < lossRate;

        nodeBehavior.receivedPacket( this, packet );
        if ( !lost && sent ) {
            monitor.reportReceived();
        } else {
            monitor.reportLost();
        }

        //Send the packet to all of our children for this stripe
        Set<Node> children = getRequiredTreeInfo( packet.getStripeId() ).getChildren();
        for ( Node child : children ) {
            //If it's not lost, would we have sent it?
            if ( !lost )
                sent &= nodeBehavior.allowForward( this, child );

            //Maybe this packet will get lost?
            if ( !lost && sent && lossRate > 0 )
                lost = sim.getRng().nextDouble() < lossRate;
            NodePacket toSend = new NodePacket( this, packet.getStripeId(), lost,
                                                sent, packet.getSize() );

            if ( sent )
                nodeBehavior.sentPacket( this, child, toSend );

            sim.scheduleEvent( sim.getTime() + calculateDelay( child, toSend ),
                               new NodeTransferEvent( toSend, child ) );
        }
    }

    /**
     * Calculates how long it takes, in seconds, to send a packet from this Node to the destination
     * Node.
     */
    private double calculateDelay( Node dest, NodePacket packet ) {
        int effectiveBW = Math.min( outboundBW, dest.inboundBW );
        return (double)( packet.getSize() * 8 ) / (double)effectiveBW;
    }

    /**
     * Returns the tree information for this Node's participation in the given stripe. The returned
     * TreeInfo may be modified.
     *
     * @param stripeId stripe to get information about.
     *
     * @return nullable TreeInfo, null if this Node is not in the stripe, or non-null otherwise.
     */
    public TreeInfo getTreeInfo( int stripeId ) {
        return trees.get( stripeId );
    }

    /**
     * Creates if needed, and returns the tree information for this Node's participation in the
     * given stripe. The returned TreeInfo may be modified. If a TreeInfo was created, then it has
     * no parent and no children.
     *
     * @param stripeId stripe to get information about
     *
     * @return non-null TreeInfo
     */
    public TreeInfo makeTreeInfo( int stripeId ) {
        TreeInfo ret = trees.get( stripeId );
        if ( ret == null ) {
            ret = new TreeInfo( stripeId );
            trees.put( stripeId, ret );
        }
        return ret;
    }

    /**
     * Resets all network information for this Node by clearing its tree information for all
     * stripes.
     */
    public void resetNetwork() {
        trees.clear();
    }

    public int getNodeId() {
        return nodeId;
    }

    public int getInboundBW() {
        return inboundBW;
    }

    public int getOutboundBW() {
        return outboundBW;
    }

    /**
     * Gets the stripe ID of the primary stripe for this Node.
     */
    public int getStripeId() {
        return stripeId;
    }

    /**
     * Returns true if Node currently receives specified stripe. A Node is a participant if its
     * makeTreeInfo method has been called for this stripe ID.
     *
     * @param stripeId stripe ID to query about
     *
     * @return true if Node is in this stripe
     */
    public boolean receivesStripe( int stripeId ) {
        return trees.containsKey( stripeId );
    }

    /**
     * Returns this Node's parent for a given stripe ID.
     *
     * @param stripeId stripe ID to query about
     *
     * @return nullable Node parent
     *
     * @throws IllegalArgumentException if this Node is not in the stripe tree
     */
    public Node getParent( int stripeId ) {
        return getRequiredTreeInfo( stripeId ).getParent();
    }

    /**
     * Returns a List of all Nodes to and including the root Node for a given stripe tree. This Node
     * is not included in the list. The immediate parent is the first element and the list continues
     * in order to the root, which is the last element.
     *
     * @param stripeId stripe ID to query about
     *
     * @return non-null List of 0 or more non-null Node
     *
     * @throws IllegalArgumentException if this Node is not in the stripe tree
     */
    public List<Node> getPathToRoot( int stripeId ) {
        List<Node> ret = new ArrayList<Node>();

        Node nextParent = this;

        while ( nextParent != null ) {
            nextParent = nextParent.getParent( stripeId );
            if ( nextParent != null )
                ret.add( nextParent );
        }

        return ret;
    }

    /**
     * Returns whether a Node is willing to accept a given child.
     *
     * @param candidateChild non-null Node
     *
     * @return true if this Node is willing to accept, false otherwise
     */
    public boolean allowChild( Node candidateChild ) {
        CheckUtils.checkNull( candidateChild, "candidateChild" );

        return nodeBehavior.allowChild( this, candidateChild ) &&
               getAvailableOutboundBW( sim.getParams() ) >= sim.getParams().getStripeBps();
    }

    /**
     * Sets the stripe ID to be the primary stripe for this Node.
     */
    public void setStripeId( int stripeId ) {
        this.stripeId = stripeId;
    }

    /**
     * Returns the TreeInfo for the given stripe.
     *
     * @return non-null TreeInfo
     *
     * @throws IllegalArgumentException if this Node is not in the stripe.
     */
    private TreeInfo getRequiredTreeInfo( int stripeId ) {
        TreeInfo ret = trees.get( stripeId );

        if ( ret == null )
            throw new IllegalArgumentException( this + " is not in stripe " + stripeId );

        return ret;
    }

    /**
     * Returns a string representation of this node: "Node [node ID in hex]"
     */
    public String toString() {
        return "Node " + Integer.toHexString( getNodeId() );
    }

    /**
     * Returns the amount of available outbound bandwidth. When this node has children in a stripe,
     * its available outbound bandwidth is decreased from the maximum.
     *
     * @param params non-null SimulationParameters (to get stripe size)
     *
     * @return amount of outbound bandwidth remaining
     */
    public int getAvailableOutboundBW( SimulationParameters params ) {
        CheckUtils.checkNull( params, "params" );

        int totalOutboundBW = 0;

        for ( Integer stripe : trees.keySet() ) {
            TreeInfo info = trees.get( stripe );
            int numChildren = info.getChildren().size();
            totalOutboundBW += numChildren * params.getStripeBps();
        }

        return ( getOutboundBW() - totalOutboundBW );
    } // getAvailableOutboundBW

    /**
     * Returns a Set of all stripe IDs this Node is currently in.
     */
    public Set<Integer> getStripes() {
        return trees.keySet();
    }

    /**
     * Returns true if this Node wants to join another stripe.
     */
    public boolean wantsJoin() {
        //trees.size() * bps compare
        int bps = sim.getParams().getStripeBps();
        boolean retVal1 = trees.size() * bps + bps <= inboundBW;
        boolean retVal2 = sim.getParams().getNumStripes() > trees.size();
        return retVal1 && retVal2;
    }
}
