package Fairness.Nodes.NodeBehaviors;

import Fairness.Nodes.Node;
import Fairness.Nodes.NodePacket;

/**
 * SimpleNodeBehavior is a NodeBehavior that always accepts children, always allows forwarding, and
 * does nothing in receivePacket. This implements a normal node with no freeloader detection.
 *
 * @author Jason Winnebeck
 */
public class SimpleNodeBehavior implements NodeBehavior {
    /**
     * Allows setting of a single behavior-specific parameter.
     *
     * @param param non-null String
     */
    public void setParameter( String param ) {
        throw new IllegalArgumentException( getClass().getName() + " does not take a parameter" );
    }

    /**
     * Asks the node behavior if the given Node is allowed to be a child of the Node this behavior
     * is for.
     *
     * @param myNode         non-null Node to which this behavior belongs.
     * @param candidateChild non-null Node that is not equal to myNode.
     *
     * @return true if the child can be added, or false otherwise.
     */
    public boolean allowChild( Node myNode, Node candidateChild ) {
        return true;
    }

    /**
     * Returns true if this node behavior is currently allowing the Node to send to the given
     * target.
     *
     * @param myNode non-null Node to which this behavior belongs.
     * @param target non-null Node that is not equal to myNode.
     */
    public boolean allowForward( Node myNode, Node target ) {
        return true;
    }

    /**
     * Event to the node behavior for when a packet is received by the Node.
     *
     * @param myNode non-null Node to which this behavior belongs.
     * @param packet non-null NodePacket destined for this behavior's Node.
     */
    public void receivedPacket( Node myNode, NodePacket packet ) {
    }

    /**
     * Event called when a Node sends a packet to a child.
     *
     * @param myNode      non-null Node to which this behavior belongs.
     * @param destination non-null Node where the packet is going.
     * @param packet      non-null NodePacket destined for this behavior's Node.
     */
    public void sentPacket( Node myNode, Node destination, NodePacket packet ) {
    }

    /**
     * Returns true if this behavior has confidence information on any Node
     */
    public boolean hasConfidenceInformation() {
        return false;
    }

    /**
     * Returns true if this behavior has debt levels for any Node
     */
    public boolean hasDebtLevelInformation() {
        return false;
    }

    /**
     * Returns this behavior's confidence of the given Node, or null if there is no opinion.
     *
     * @param n non-null Node to get information about
     *
     * @return nullable Double
     */
    public Double getConfidence( Node n ) {
        return null;
    }

    /**
     * Returns this behavior's debt level measurement of the given Node, or null if there is no
     * opinion.
     *
     * @param n non-null Node to get information about
     *
     * @return nullable Double
     */
    public Double getDebtLevel( Node n ) {
        return null;
    }
}
