package Fairness;

import Fairness.Graphing.Graph;
import Fairness.Graphing.GraphBounds;
import Fairness.Graphing.NamedListXYSeries;
import Fairness.Nodes.Node;
import Fairness.Nodes.NodeBehaviors.NodeBehavior;
import org.gillius.Util.CheckUtils;

import java.io.File;
import java.io.IOException;
import java.io.PrintWriter;
import java.util.ArrayList;
import java.util.Collection;
import java.util.Collections;
import java.util.HashMap;
import java.util.Iterator;
import java.util.List;
import java.util.Map;

/**
 * SimulationRunner wraps a Simulation and runs it, generating the output files and writing them to
 * disk after the simulation is complete.
 *
 * @author Jason Winnebeck
 */
public class SimulationRunner {
    private static final String DELIM = ", ";

    private final List<NodeMonitor> monitors;
    private final Collection<NodeBehavior> behaviors;
    private final Map<String, List<Node>> nodeGroups;

    private final Simulation sim;
    private final File outputDirectory;

    /**
     * Constructs a new SimulationRunner. The number and format of output files is described in the
     * user guide.
     *
     * @param monitors        non-null List of all node monitors used in the simulation
     * @param behaviors       non-null Collection of all NodeBehaviors in the sim
     * @param nodeGroups      non-null Map to group Nodes into groups by name
     * @param sim             non-null Simulation to run
     * @param outputDirectory non-null File denoting an existing directory for output
     */
    public SimulationRunner( List<NodeMonitor> monitors, Collection<NodeBehavior> behaviors,
                             Map<String, List<Node>> nodeGroups, Simulation sim,
                             File outputDirectory ) {
        CheckUtils.checkNull( monitors, "monitors" );
        CheckUtils.checkNull( behaviors, "behaviors" );
        CheckUtils.checkNull( sim, "sim" );
        CheckUtils.checkNull( outputDirectory, "outputDirectory" );

        if ( !outputDirectory.isDirectory() ) {
            throw new IllegalArgumentException( "outputDirectory must be a directory" );
        }

        this.monitors = monitors;
        this.behaviors = behaviors;
        this.nodeGroups = nodeGroups;
        this.sim = sim;
        this.outputDirectory = outputDirectory;
    }

    /**
     * Runs the simulation and generates output files.
     *
     * @throws SimulationException if the Simulation could not finish or errors occured writing
     *                             output files.
     */
    public void runSimulation() throws SimulationException {
        while ( sim.getTime() < sim.getParams().getSimDuration() ) {
            if ( !sim.next() ) {
                throw new SimulationException( "Not enough events to run the simulation " +
                                               "to completion" );
            }
        }
        System.out.println( "  Simulation complete, writing output" );

        //Output packet received information from our monitors
        for ( NodeMonitor monitor : monitors ) {
            writeOutput( monitor );
        }

        writeCombinedReceived();

        boolean hasConf = false;
        boolean hasDebt = false;

        for ( Iterator it = behaviors.iterator(); it.hasNext() && ( !hasConf || !hasDebt ); ) {
            NodeBehavior behavior = (NodeBehavior) it.next();
            hasConf |= behavior.hasConfidenceInformation();
            hasDebt |= behavior.hasDebtLevelInformation();
        }

        if ( hasConf )
            writeConfidenceCDF();

        if ( hasDebt )
            writeDebtCDF();
    }

    /**
     * Writes the ancestor rating confidence CDF. The precondition for this method is that there
     * are node behaviors with confidence in the system.
     */
    private void writeConfidenceCDF() throws SimulationException {
        Map<Node, Double> values = getAveragesFromBehavior( QueryType.Confidence );

        double min = Collections.min( values.values() );
        double max = Collections.max( values.values() );

        //Now, calculate CDF by group
        File dataFile = new File( outputDirectory, sim.getName() + "-ConfCDF.csv" );
        File imageFile = new File( outputDirectory, sim.getName() + "-ConfCDF.png" );

        createCDF( dataFile, imageFile, values, (max-min)/50, min, max,
                   "CDF of Confidence Level", "Confidence" );
    }

    /**
     * Writes the debt level CDF. The precondition for this method is that there
     * are node behaviors with debt levels in the system.
     */
    private void writeDebtCDF() throws SimulationException {
        Map<Node, Double> values = getAveragesFromBehavior( QueryType.Debt );

        double min = Collections.min( values.values() );
        double max = Collections.max( values.values() );

        //Now, calculate CDF by group
        File dataFile = new File( outputDirectory, sim.getName() + "-DebtCDF.csv" );
        File imageFile = new File( outputDirectory, sim.getName() + "-DebtCDF.png" );

        createCDF( dataFile, imageFile, values, (max-min)/50, min, max,
                   "CDF of Debt Level", "Debt Level" );
    }

    /**
     * An enumeration to get the type of query made on a NodeBehavior.
     */
    private static enum QueryType {
        Debt, Confidence
    }

    /**
     * Returns an average rating for each Node in the network by finding the average of all Nodes'
     * opinion of a single Node.
     *
     * @param type type of query, debt level or confidence rating
     *
     * @return non-null Map of each Node in the system and the average rating
     */
    private Map<Node, Double> getAveragesFromBehavior( QueryType type ) {
        Map<Node, Double> ret = new HashMap<Node, Double>();

        //Calculate average debt levels for all nodes
        for ( List<Node> nodeList : nodeGroups.values() ) {
            for ( Node node : nodeList ) {
                int count = 0;
                double value = 0.0;

                //Query each behavior for its view of node's debt
                for ( NodeBehavior behavior : behaviors ) {
                    Double val;
                    switch ( type ) {
                        case Debt:
                            val = behavior.getDebtLevel( node );
                            break;
                        case Confidence:
                            val = behavior.getConfidence( node );
                            break;
                        default:
                            throw new AssertionError( "Unknown QueryType " + type );
                    }
                    if ( val != null ) {
                        ++count;
                        value += val;
                    }
                }

                //Find the average, if there were any debts
                if ( count > 0 ) {
                    if ( ret.containsKey( node ) )
                        throw new AssertionError( "Duplcate node?" );

                    ret.put( node, value / count );
                }
            }
        }

        return ret;
    }

    /**
     * Creates a CDF graph and writes it to CSV and PNG files.
     *
     * @param dataFile     output CSV file location
     * @param imageFile    output PNG file location
     * @param ratings      ratings for each node
     * @param sampleStride interval between each CDF sample
     * @param startValue   start of sampling interval
     * @param endValue     end of sampling interval
     * @param graphTitle   title of graph
     * @param graphBottom  x-axis label
     *
     * @throws SimulationException if an error occurs in writing the graph, including IO errors.
     */
    private void createCDF( File dataFile, File imageFile, Map<Node, Double> ratings,
                            double sampleStride, double startValue, double endValue,
                            String graphTitle,
                            String graphBottom ) throws SimulationException {
        List<NamedListXYSeries> series = new ArrayList<NamedListXYSeries>( nodeGroups.size() );

        try {
            PrintWriter out = new PrintWriter( dataFile );
            out.print( "Threshold" );
            //iterate using entrySet to make sure we get the same order as later
            for ( Map.Entry<String, List<Node>> group : nodeGroups.entrySet() ) {
                out.print( DELIM );
                out.print( group.getKey() );
                series.add( new NamedListXYSeries( group.getKey() ) );
            }
            out.println();

            for ( double currThresh = startValue; currThresh <= endValue; currThresh += sampleStride ) {
                out.print( currThresh );
                int seriesIdx = 0;
                for ( Map.Entry<String, List<Node>> group : nodeGroups.entrySet() ) {
                    int totalNodes = 0;
                    int nodesUnder = 0;

                    for ( Node node : group.getValue() ) {
                        if ( ratings.containsKey( node ) ) {
                            ++totalNodes;
                            if ( ratings.get( node ) <= currThresh )
                                ++nodesUnder;
                        }
                    }

                    double percent;
                    if ( totalNodes == 0 )
                        percent = 0.0;
                    else
                        percent = (double)nodesUnder / totalNodes;

                    out.print( DELIM );
                    out.print( percent );

                    series.get( seriesIdx ).add( currThresh, percent );
                    ++seriesIdx;
                }
                out.println();
            }
            out.close();
            if ( out.checkError() )
                throw new SimulationException( "PrintStream reported formatting error" );

        } catch ( Exception e ) {
            throw new SimulationException( "Error writing data file " +
                                           dataFile.getAbsoluteFile(), e );
        }

        GraphBounds bounds = new GraphBounds( startValue, endValue, 0.0, 1.0 );
        Graph graph = new Graph( graphTitle, "Percent", graphBottom, series, bounds );
        writeGraph( graph, imageFile );
    }

    /**
     * Constructs the data file and graph for the combined packets received graph.
     *
     * @throws SimulationException if an error occurs in writing the graph, including IO errors.
     */
    private void writeCombinedReceived() throws SimulationException {
        if ( monitors.size() == 0 )
            return;
        int length = monitors.get( 0 ).getMarks().size();
        List<NamedListXYSeries> series = new ArrayList<NamedListXYSeries>( monitors.size() );
        for ( NodeMonitor monitor : monitors ) {
            series.add( new NamedListXYSeries( monitor.getName() ) );
        }

        File dataFile = new File( outputDirectory, sim.getName() + "-combined.csv" );
        File imageFile = new File( outputDirectory, sim.getName() + "-combined.png" );

//        if ( dataFile.exists() )
//            throw new SimulationException( dataFile.getAbsoluteFile() + " already exists" );
//        if ( imageFile.exists() )
//            throw new SimulationException( imageFile.getAbsoluteFile() + " already exists" );

        try {
            //Print the header line
            PrintWriter out = new PrintWriter( dataFile );
            out.print( "Rebuild" );
            for ( NodeMonitor monitor : monitors ) {
                out.print( DELIM );
                out.print( monitor.getName() );
                if ( monitor.getMarks().size() != length )
                    throw new SimulationException( "monitors have different mark lengths" );
            }
            out.println( ", Total" );

            //Print each of the rows for each mark
            for ( int i = 0; i < length; ++i ) {
                double total = 0.0;
                out.print( i );
                int monNum = 0;
                for ( NodeMonitor monitor : monitors ) {
                    out.print( DELIM );
                    double p = monitor.getMarks().get( i ).getReceivedPercent();
                    out.print( p );
                    total += p;
                    series.get( monNum ).add( i, p );
                    monNum++;
                }
                out.print( DELIM );
                out.println( total );
            }

            out.close();
            if ( out.checkError() )
                throw new SimulationException( "PrintStream reported formatting error" );

        } catch ( Exception e ) {
            throw new SimulationException( "Error writing data file " +
                                           dataFile.getAbsoluteFile(), e );
        }

        GraphBounds bounds = new GraphBounds( 0, length-1, 0.0, 1.0 );
        Graph graph = new Graph( "Received Packets", "Percent", "Reconstructions", series, bounds );
        writeGraph( graph, imageFile );
    }

    /**
     * Writes the output for a single NodeMonitor to a file named after the simulation and the
     * NodeMonitor.
     *
     * @param monitor non-null NodeMonitor
     *
     * @throws SimulationException if an error occurs in writing the graph, including IO errors.
     */
    private void writeOutput( NodeMonitor monitor ) throws SimulationException {
        final String name = monitor.getName();

        File dataFile  = new File( outputDirectory, sim.getName() + "-" + name + ".csv" );
        File imageFile = new File( outputDirectory, sim.getName() + "-" + name + ".png" );

//        if ( dataFile.exists() )
//            throw new SimulationException( dataFile.getAbsoluteFile() + " already exists" );
//        if ( imageFile.exists() )
//            throw new SimulationException( imageFile.getAbsoluteFile() + " already exists" );

        //Write all of the marks for the NodeMonitor to the CSV file
        int i = 0;
        try {
            PrintWriter out = new PrintWriter( dataFile );
            out.println( "Rebuild, Total, Lost, Lost%, Received, Received%" );
            for ( NodeMonitor.Mark mark : monitor.getMarks() ) {
                out.println( i + DELIM + mark.getTotal() + DELIM +
                             mark.getLost() + DELIM + mark.getLostPercent() + DELIM +
                             mark.getReceived() + DELIM + mark.getReceivedPercent() );
                ++i;
            }

            out.close();
            if ( out.checkError() )
                throw new SimulationException( "PrintStream reported formatting error" );

        } catch ( Exception e ) {
            throw new SimulationException( "Error writing data file " +
                                           dataFile.getAbsoluteFile(), e );
        }

        //Now we construct the graph for this monitor
        NamedListXYSeries data = new NamedListXYSeries( name );
        i = 0;
        for ( NodeMonitor.Mark mark : monitor.getMarks() ) {
            data.add( i++, mark.getReceivedPercent() );
        }

        List<NamedListXYSeries> seriesList = Collections.singletonList( data );
        GraphBounds bounds = GraphBounds.calculateBounds( seriesList );
        bounds.setYBottom( 0.0 );
        bounds.setYTop( 1.0 );
        Graph graph = new Graph( "Received Packets for " + name, "Percent",
                                 "Reconstructions", seriesList, bounds );
        writeGraph( graph, imageFile );
    }

    /**
     * Writes a Graph to a File, converting an IOException to a SimulationException.
     */
    private static void writeGraph( Graph graph, File imageFile ) throws SimulationException {
        try {
            graph.writeColorPng( imageFile );
        } catch ( IOException e ) {
            throw new SimulationException( "Error writing graph image", e );
        }
    }
}
