package Fairness.Trees;

import Fairness.Nodes.Node;
import Fairness.Nodes.TreeInfo;
import Fairness.NumberUtil;

import java.util.ArrayList;
import java.util.Collection;
import java.util.Collections;
import java.util.HashSet;
import java.util.Iterator;
import java.util.List;
import java.util.Set;
import java.util.TreeSet;


/**
 * PastryTreeManager implements the Pastry-style tree routing algorithm from the SplitStream
 * paper. The algorithm is covered in detail in the paper analysis in the final report for the
 * SplitStream.
 *
 * @author John Kaeuper
 */
public class PastryTreeManager extends TreeManager {
    private final Set<Node> spareCG = new HashSet<Node>();

    private final Set<Integer> allocatedIds = new TreeSet<Integer>();

    private final List<Integer> stripeIds = new ArrayList<Integer>();

    private final Set<Node> rootNodes = new HashSet<Node>();

    @Override protected void setUp() {
        super.setUp();
        // now, assign the stripeId's based on how many stripes there
        // are (note that no Nodes will have NodeId's equal to these stripeIds)
        for ( int i = 0; i < sim.getParams().getNumStripes(); ++i )
            stripeIds.add( getNextNodeId() );
    }

    /**
     * After a tree reconstruction, returns the nodes that are the roots of their stripe trees. It
     * is assumed that the Node is the root of its primary stripe and of no other. That is, Node
     * getParent( getStripeId() ) should return null.
     *
     * @return non-null Collection of 0 or more Node.
     */
    public Collection<Node> getRootNodes() {

        return rootNodes;

    }

    /**
     * Adds a Node to this TreeManager for it to manage and build into trees. This method exists
     * just for extra error checking purposes.
     *
     * @param node non-null Node that has an ID that came from getNextNodeId.
     */
    public void addNode( Node node ) {
        if ( !allocatedIds.contains( node.getNodeId() ) )
            throw new IllegalArgumentException(
                node + " did not get its ID from this tree manager." );

        super.addNode( node );
    }

    /**
     * Returns the next node ID using the node ID allocation algorithm from this TreeManager.
     */
    public int getNextNodeId() {

        int randFirstDigit =
            sim.getRng().nextInt( sim.getParams().getNumStripes() );
        return getNextNodeId( randFirstDigit );

    }

    /**
     * A method to completely rebuild the Node tree given all nodes in the system.  The Pastry
     * routing algorithm is used to form the parent-child relationships, and the Scribe group
     * mechanisms are used to maintain the interior-node-disjoint property of the forest.
     */
    public void rebuildTree() {
        for ( Node node : getAllNodes() ) {
            node.resetNetwork();
        }

        // now that Nodes are reset, we must init the rest of their
        // state again, before building the forest
        initNodes();

        // Note: make sure that findParent is called in randomized order
        // of Nodes each time this method is called

        List<Node> allNodes = new ArrayList<Node>( getAllNodes() );


        Collections.shuffle( allNodes, sim.getRng() );

        // for each Node in the system, find a parent for each stripe that the
        // node has inbound bandwidth capacity to receive
        for ( Node node : allNodes ) {

            int numStripesRecv = node.getInboundBW() /
                                 sim.getParams().getStripeBps();

            // this assumes Node doesn't care which stripes it receives,
            // as long as it receives numStripesRecv of them; that is,
            // all stripes are considered to be the same, so Node just
            // receives random stripes from among stripeIds
            List<Integer> stripesToReceive =
                new ArrayList<Integer>( stripeIds );
            for ( int i = 0; i < numStripesRecv; ++i ) {

                int randIndex = sim.getRng().nextInt(stripesToReceive.size());
                findParent( stripeIds.get( randIndex ), node );
                stripesToReceive.remove( stripeIds.get( randIndex ) );

            }

        }

    }

    /**
     * Finds the parent for the given node for a particular stripe identified by the
     * stripeId.
     *
     * @param n non-null Node that has an ID that came from getNextNodeId.
     * @param stripeId - the int id identifying the stripe the node is attempting to receive
     */
    public void findParent( int stripeId, Node n ) {

        int curId = n.getNodeId();
        List<Node> pathToRoot = new ArrayList<Node>();
        pathToRoot.add( n );

        Collection<Node> theNodes = getAllNodes();

        int prefixMatchSize = NumberUtil.numMatchingPrefix( curId, stripeId );
        int minDifference = curId - stripeId;
        if ( minDifference < 0 )
            minDifference = -1 * minDifference;
        boolean pathCompleted = false;

        while ( !pathCompleted ) {

            Node nextPathNode = null;

            Iterator<Node> iter = theNodes.iterator();
            boolean found = false;

            while ( iter.hasNext() && !found ) {
                Node nextNode = iter.next();
                int nextId = nextNode.getNodeId();
                int newPrefixSize =
                    NumberUtil.numMatchingPrefix( nextId, stripeId );
                if ( newPrefixSize > prefixMatchSize ) {
                    prefixMatchSize = newPrefixSize;
                    nextPathNode = nextNode;
                    found = true;
                } else if ( newPrefixSize == prefixMatchSize ) {
                    int curDifference = nextId - stripeId;
                    if ( curDifference < 0 )
                        curDifference = -1 * curDifference;
                    if ( curDifference < minDifference ) {
                        minDifference = curDifference;
                        nextPathNode = nextNode;
                    }
                }
            } // while going through Nodes

            if ( nextPathNode != null )
                pathToRoot.add( nextPathNode );
            else
                pathCompleted = true;

        } // while not reached root of stripe tree

        // DISCOVER A ROOT NODE
        //
        // now the final Node in pathToRoot list is the root of a stripe
        // tree - so add it to rootNodes Set
        if ( rootNodes.size() < sim.getParams().getNumStripes() )
            rootNodes.add( pathToRoot.get( pathToRoot.size() - 1 ) );
        //
        // DONE DISCOVERING A ROOT NODE


        List<Boolean> childPermissions = new ArrayList<Boolean>();

        // update TreeInfo's of each Node on path to root
        for ( int i = 0; i < pathToRoot.size() - 1; ++i ) {

            Node node1 = pathToRoot.get( i );
            Node node2 = pathToRoot.get( i + 1 );

            // TODO: we have to pass a Node here, or maybe we need a
            // wantsMoreChildren method?
            // perhaps want to use TreeBehavior wantJoinMore method?
            if ( !node2.allowChild( node1 ) )
                childPermissions.add( false );
            else
                childPermissions.add( true );


            node2.getTreeInfo( stripeId ).addChild( node1 );
            node1.getTreeInfo( stripeId ).setParent( node2 );
        }

        // update spare capacity group
        for ( int i = 1; i < pathToRoot.size(); ++i ) {

            Node curNode = pathToRoot.get( i );
            if ( curNode.getAvailableOutboundBW( sim.getParams() ) <= 0 )
                spareCG.remove( curNode );
        }

        System.err.println( "about to orphan if necessary" );
        // now orphan children if necessary
        for ( int i = 1; i < pathToRoot.size(); ++i ) {

            Node parent = pathToRoot.get( i );
            Node child = pathToRoot.get( i - 1 );

            if ( !childPermissions.get( i - 1 ) )
                orphanChild( parent, child, stripeId );
        }

        System.err.println( "past orphaning - findParent returned" );

    } // findParent

    /**
     * Determines the orphan that will be rejected by n (could be justAdded) and
     * then finds a new parent for this orphan for the stripe identified by the
     * stripeOrphanWants id.
     *
     * @param n non-null Node that has an ID that came from getNextNodeId.
     * @param justAdded the Node that was just added as the child of n
     * @param stripeOrphanWants the id of the stripe that the orphan needs
     */
    public void orphanChild( Node n, Node justAdded, int stripeOrphanWants ) {

        System.err.println( "orphanChild called" );
        // detemine the orphan to be rejected
        Node orphan = determineOrphan( n, justAdded );

        // find a new parent for the orphan
        findOrphanParent( n, orphan, stripeOrphanWants );

    }

    /**
     * This method finds a parent for orphan for the specified stripe by searching
     * randomly in the "Spare Capacity Group" (this terminology is borrowed from the
     * SplitStream implementation, but in this implementation, the Spare Capacity Group is
     * nothing more than a Set of nodes having spare outbound bandwidth enough to handle
     * one more child.
     *
     * @param orphan the orphan who needs to receive the stripe identified by stripeId
     * @param stripeId the integer id for the stripe that orphan needs
     */
    public void searchSCG( Node orphan, int stripeId ) {

        System.err.println( "searchSCG called" );

        // make copy of spare capacity group and get a random Node from it
        Set<Node> spareCopy = new HashSet<Node>( spareCG );
        Node[] spareArray = spareCopy.toArray( new Node[spareCopy.size()] );
        Node n = spareArray[sim.getRng().nextInt( spareArray.length )];

        boolean done = false;

        while ( !done ) {

            if ( spareCopy.size() > 1 ) {

                // if n receives needed stripe and orphan is not in path
                // to root (avoid cycles!)
                if ( n.receivesStripe( stripeId ) &&
                     !n.getPathToRoot( stripeId ).contains( orphan ) ) {

                    n.getTreeInfo( stripeId ).addChild( orphan );
                    orphan.getTreeInfo( stripeId ).setParent( n );

                    if ( n.getAvailableOutboundBW( sim.getParams() ) <= 0 )
                        spareCG.remove( n );

                    done = true;

                // otherwise, get a new Random node from SCG + continue
                } else {
                    spareCopy.remove( n );
                    Node[] newSpareArray =
                        spareCopy.toArray( new Node[spareCopy.size()] );
               n = newSpareArray[sim.getRng().nextInt( newSpareArray.length )];
                }

            // if only one node left, this is our last chance.....
            } else if ( spareCopy.size() == 1 ) {

                if ( n.receivesStripe( stripeId ) &&
                     !n.getPathToRoot( stripeId ).contains( orphan ) ) {

                    n.getTreeInfo( stripeId ).addChild( orphan );
                    orphan.getTreeInfo( stripeId ).setParent( n );

                    if ( n.getAvailableOutboundBW( sim.getParams() ) <= 0 )
                        spareCG.remove( n );

                    done = true;

                // since we're in our last chance + the only trouble is
                // that we'd have cycle if added orphan, then we do this
                // funky thing where orphan just replaces a leaf Node and
                // displaced leaf does a recursive searchSCG call.....
                } else if ( n.receivesStripe( stripeId ) ) {

                    Node leaf = null;

                    Collection<Node> allNodes = getAllNodes();
                    Iterator<Node> allIter = allNodes.iterator();

                    // find a leaf node in desired stripe tree that is
                    // not a descendant of the orphan
                    while ( allIter.hasNext() ) {
                        Node nextNode = allIter.next();
                        if ( nextNode.receivesStripe( stripeId ) &&
                             nextNode.getTreeInfo( stripeId ).getChildren().isEmpty() &&
                             !nextNode.getPathToRoot( stripeId ).contains( orphan ) ) {
                            leaf = nextNode;
                            break;
                        }
                    }

                    Node leafParent = leaf.getTreeInfo( stripeId ).getParent();
                    leafParent.getTreeInfo( stripeId ).removeChild( leaf );
                    leafParent.getTreeInfo( stripeId ).addChild( orphan );
                    orphan.getTreeInfo( stripeId ).setParent( leafParent );

                    // should we set leaf's parent to be null before
                    // we do recursive searchSCG call (which should find
                    // a new parent for leaf, but not immediately)?

                    // find a new parent for leaf that was orphaned
                    searchSCG( leaf, stripeId );
                } else {
                    System.err.println( "Orphaned Node - " + orphan +
                                        " - cannot receive stripe for " +
                                        "stripeId: " +
                                        Integer.toHexString( stripeId ) );
                    System.exit( 1 );
                }
            } else if ( spareCopy.isEmpty() ) {

                System.err.println( "Orphaned Node - " + orphan +
                                    " - cannot receive stripe for " +
                                    "stripeId: " +
                                    Integer.toHexString( stripeId ) );
                System.exit( 1 );
            }
        }
    }

    /**
     * This is a private method called by orphanChild which determines which child
     * of n will be selected as the orphan to reject (this could be the node justAdded).
     *
     * @param n the method selects an orphan from among the children of this node
     * @param justAdded the node that was just added as the child of node n
     * @return the node that is selected to be orphaned from n
     */
    private Node determineOrphan( Node n, Node justAdded ) {

        Node orphan;

        // get Set of all n's children (all children, not just children
        // in a single stripe tree)
        Set<Node> nChildren = new HashSet<Node>();

        Set<Integer> stripes = n.getStripes();

        Iterator<Integer> stripeIter = stripes.iterator();
        while ( stripeIter.hasNext() ) {
            TreeInfo curTreeInfo = n.getTreeInfo( stripeIter.next() );
            nChildren.addAll( curTreeInfo.getChildren() );
        }

        int stripeId = n.getStripeId();  // n's primary stripe

        // get the Set of children for which these children are in a tree
        // whose stripeId shares no prefix with n's nodeId
        Set<Node> goodChildren = n.getTreeInfo( stripeId ).getChildren();
        Set<Node> badChildren = new HashSet<Node>( nChildren );
        badChildren.removeAll( goodChildren );


        if ( !badChildren.isEmpty() ) {
            if ( badChildren.contains( justAdded ) )
                orphan = justAdded;
            else {
                Node[] badArray =
                    badChildren.toArray( new Node[badChildren.size()] );
                orphan = badArray[sim.getRng().nextInt( badArray.length )];
            }
        } else {
            Set<Node> shortPrefix = new HashSet<Node>();
            int minPrefixSize = Integer.MAX_VALUE;

            Iterator<Node> childIter = nChildren.iterator();
            while ( childIter.hasNext() ) {
                Node child = childIter.next();
                int childId = child.getNodeId();
                int curPrefixSize =
                    NumberUtil.numMatchingPrefix( childId, stripeId );
                if ( curPrefixSize < minPrefixSize ) {
                    minPrefixSize = curPrefixSize;
                    shortPrefix = new HashSet<Node>();
                    shortPrefix.add( child );
                } else if ( curPrefixSize == minPrefixSize )
                    shortPrefix.add( child );
            }

            if ( shortPrefix.contains( justAdded ) )
                orphan = justAdded;
            else {
                Node[] shortArray =
                    shortPrefix.toArray( new Node[shortPrefix.size()] );
                orphan =
                    shortArray[sim.getRng().nextInt( shortArray.length )];
            }
        }

        return orphan;

    } // determineOrphan (private)

    /**
     * This is a private method called by orphanChild that finds a new parent for the orphan
     * that was just rejected as a child of n by the determineOrphan method.  The mechanism for
     * finding this parent is not the same as in the findParent method.  The mechanism used
     * is based on the Scribe "push-down" method described in the research paper on SplitStream
     * (for an analysis of this paper, please refer to the final report on the project website).
     *
     * @param n the former parent of orphan
     * @param orphan the orphan that was selected by determineOrphan to be rejected from n
     * @param stripeOrphanWants the identifier for the stripe that orphan needs
     */
    private void findOrphanParent( Node n, Node orphan,
                                   int stripeOrphanWants ) {

        // first, find stripeId of orphan and remove it from n's children
        // in that stripe tree (and update spare capacity group if needed)
        TreeInfo parentTreeInfo = n.getTreeInfo( stripeOrphanWants );
        parentTreeInfo.removeChild( orphan );
        if ( n.getAvailableOutboundBW( sim.getParams() ) > 0 )
            spareCG.add( n );

        // should we set orphan's parent to null before finding new
        // parent here below?
        Set<Node> orphanSiblings = parentTreeInfo.getChildren();
        if ( orphanSiblings.isEmpty() )
            searchSCG( orphan, stripeOrphanWants );
        else {
            Set<Node> goodPrefix = new HashSet<Node>();
            Iterator<Node> siblingIter = orphanSiblings.iterator();
            while ( siblingIter.hasNext() ) {
                Node sibling = siblingIter.next();
                if ( sibling.getStripeId() == stripeOrphanWants )
                    goodPrefix.add( sibling );
            }

            if ( goodPrefix.isEmpty() )
                searchSCG( orphan, stripeOrphanWants );
            else {
                Node newParent;
                Node[] goodArray =
                    goodPrefix.toArray( new Node[goodPrefix.size()] );
                newParent =
                    goodArray[sim.getRng().nextInt( goodArray.length )];

                boolean willTakeOrphan = false;
                if ( newParent.allowChild( orphan ) )
                    willTakeOrphan = true;
                else
                    System.err.println( "allowChild returned false" );

                newParent.getTreeInfo( stripeOrphanWants ).addChild( orphan );
                orphan.getTreeInfo( stripeOrphanWants ).setParent( newParent );

                // update spare capacity group
                if ( newParent.getAvailableOutboundBW(
                     sim.getParams() ) <= 0 )
                    spareCG.remove( newParent );

                // orphan another child if necessary
                if ( !willTakeOrphan ) {
                    System.err.println( "orphanChild recursively called" );
                    orphanChild( newParent, orphan, stripeOrphanWants );

                }
            }
        }

    } // findOrphanParent (private)

    /**
     * This method is for initializing the rest of the Nodes' states after
     * their nodeIds have been assigned and they've been added to the
     * nodes Set of this PastryTreeManager.  The nodes primary stripe id's
     * are assigned, their TreeInfo's are set up for each stripe, and they
     * are added to the spareCG set.
     */
    private void initNodes() {

        Collection<Node> theNodes = getAllNodes();
        Iterator<Node> nodeIter = theNodes.iterator();

        while ( nodeIter.hasNext() ) {
            Node curNode = nodeIter.next();

            // set the Node primary stripeIds based on their id's - that is,
            // a Node's primary stripeId is the stripeId that shares a
            // prefix with its nodeId
            for ( int i = 0; i < stripeIds.size(); ++i ) {
                int stripe = stripeIds.get( i );

                // set up the TreeInfo for curNode for each stripe
                curNode.makeTreeInfo( stripe );

                if ( curNode.getNodeId() == -2119832571 &&
                     stripe == -1747611997 )
                    System.out.println( "made a tree info for that node2 " +
                                        "and did for that stripe!" );




                // set each Node's primary stripeId
                if ( NumberUtil.getDigit( stripe, 0 ) ==
                     NumberUtil.getDigit( curNode.getNodeId(), 0 ) )
                    curNode.setStripeId( stripe );

            }

            // initially, all Nodes are in the SCG since no parent-child
            // relationships have been established yet
            spareCG.add( curNode );

        } // done assigning stripeIds and TreeInfos

    }


    /**
     * Returns the next node ID using the node ID allocation algorithm from
     * this TreeManager.  This uses PastryTreeManager's particular algorithm
     * that needs to know the first digit of the nodeId.  This first digit will
     * determine which stripe the node assigned this id will be required to forward.
     * (Please see the discussion of the Scribe group communication system in the
     * final report on the project website for more detail).
     *
     * @param firstDigit - the random nodeId's firstDigit
     */
    private int getNextNodeId( int firstDigit ) {

        int ret;

        // TODO: Do we need to adjust the random generation because of the
        // stripe IDs?

        do {

           ret = NumberUtil.getRandomWithPrefix( firstDigit, 1, sim.getRng() );

        } while ( allocatedIds.contains( ret ) );

        allocatedIds.add( ret );

        return ret;

    }

} // PastryTreeManager
