package Fairness.Trees;

import Fairness.Nodes.Node;

import java.util.ArrayList;
import java.util.Collection;
import java.util.Collections;
import java.util.List;
import java.util.Map;
import java.util.TreeMap;

/**
 * RandomTreeManager constructs a tree randomly by giving each node a single turn to join a tree.
 * The manager searches the nodes in the desired stripe tree randomly by visiting nodes in that stripe
 * tree until it finds an appropriate parent for the joining node.
 *
 * @author Jason Winnebeck
 */
public class RandomTreeManager extends TreeManager {
    private int nodeIdCounter = 0;

    private Map<Integer, Node> stripeRoots = new TreeMap<Integer, Node>();

    private List<Node> inNetwork = new ArrayList<Node>();

    /**
     * After a tree reconstruction, returns the nodes that are the roots of their stripe trees. It
     * is assumed that the Node is the root of its primary stripe and of no other. That is, Node
     * getParent( getStripeId() ) should return null.
     *
     * @return non-null Collection of 0 or more Node.
     */
    public Collection<Node> getRootNodes() {
        return stripeRoots.values();
    }

    /**
     * A method to completely rebuild the Node tree given all nodes in the system.
     */
    public void rebuildTree() {
        //get all nodes and shuffle
        List<Node> allNodes = new ArrayList<Node>( getAllNodes() );

        inNetwork.clear();
        for ( Node node : allNodes ) {
            node.resetNetwork();
        }

        // assign random primary stripeId's to each Node
        for ( Node node : allNodes ) {

            int stripe = sim.getRng().nextInt( sim.getParams().getNumStripes() );

            node.setStripeId( stripe );
//            System.err.println( "setting Node stripeId: " + stripe );
        }

        boolean quit = false;
        while ( !quit ) {
            boolean anyJoined = false;
            boolean anyWant = false;

            // shuffle the list
            Collections.shuffle( allNodes, sim.getRng() );

            for ( Node node : allNodes ) {

                // Does the node want to join another stripe
                boolean wantJoin = node.wantsJoin();
                //anyWant |= wantJoin;
                if ( wantJoin ) {
                    anyWant = true;
//                    System.err.println( "Node wants to join: " + node.getNodeId() );
                }
                // if yes, findParent
                if ( wantJoin ) {
                    //anyJoined |= findParent( node );
                    if ( findParent( node ) )
                        anyJoined = true;
                }
            }

            //quit is true if no one wants to join a stripe, OR if nodes want to join, but no one
            //actually DID join -- this is an error case
            boolean networkFailure = false;
            if ( anyWant && !anyJoined ) {
//                System.err.println( "One or more Nodes were unable to " +
//                                    "receive a stripe they wanted" );
                networkFailure = true;
            }

            if ( !networkFailure )
                quit = !anyWant;
            else
                quit = true;
        }
    }

    private boolean findParent( Node node ) {

//        System.err.println( "entered findParent" );

        int stripeToJoin;

        // if the node is not in ANY stripes, join primary stripe, else
        if ( node.getStripes().isEmpty() ) {
            stripeToJoin = node.getStripeId();
//            System.err.println( "Node will try to get its primary stripe, which " +
//                                "is: " + stripeToJoin );

        // decide at random what stripe to join among stripes that
        // node does not yet receive
        } else {
            do {
                stripeToJoin =
                    sim.getRng().nextInt( sim.getParams().getNumStripes() );
            } while ( node.getStripes().contains( stripeToJoin ) );

//            System.err.println("Node is trying receive a random stripe....., " +
//                               "which is: " + stripeToJoin );


        }

        // see if no nodes in network yet receive stripeToJoin - if none
        // receive it, then this node being added will become the root node
        boolean stripeTreeIsEmpty = true;
        for ( Node networkNode : inNetwork ) {
            if ( networkNode.receivesStripe( stripeToJoin ) ) {
                stripeTreeIsEmpty = false;
                break;
            }

        }

        if ( stripeTreeIsEmpty ) {
//            System.err.println( "the stripe tree is empty...." );
            if ( !stripeRoots.containsValue( node ) ) {
//                System.err.println( "Node is not already a root in another " +
//                                    "tree, so we will make it the root of this " +
//                                    "stripe tree, and set it's parent to null and" +
//                                    " add it to inNetwork if not already there " +
//                                    "and findParent returns true" );
                if ( !inNetwork.contains( node ) )
                    inNetwork.add( node );
                stripeRoots.put( stripeToJoin, node );
                node.makeTreeInfo( stripeToJoin ).setParent( null );
                return true;
            } else {
//                System.err.println( "Node is already a root elsewhere, so " +
//                                    "findParent returns false....." );
                return false;
            }
        } else {
//            System.err.println( "the stripe tree is NOT empty....." );
        }

        //Iterate randomly over all nodes currently in the network
        Collections.shuffle( inNetwork, sim.getRng() );

        // Pass 1, try to find parent that is supposed to forward
        // stripeToJoin and can accept node as a child
        boolean joined = false;
//        System.err.println( "Node is trying to find a GOOD parent...." );
        for ( Node possibleParent : inNetwork ) {

            if ( possibleParent != node &&
                 possibleParent.getStripeId() == stripeToJoin &&
                 possibleParent.receivesStripe( stripeToJoin ) &&
                 possibleParent.allowChild( node ) ) {
                possibleParent.getTreeInfo( stripeToJoin ).addChild( node );
                node.makeTreeInfo( stripeToJoin ).setParent( possibleParent );
                joined = true;
//                System.err.println( "Node found a GOOD parent! Parent is: " +
//                                    possibleParent.getNodeId() );
                break;
            }
        }

        if ( !joined ) {
//            System.err.println( "Node is trying to find a BAD parent...." );
             // if Pass 1 failde, try Pass 2 - try SCG - try to find parent
            // that forwards stripeToJoin even though it is not supposed to
            // forward stripeToJoin and see if it can accept node as a child
            for ( Node possibleParent : inNetwork ) {
//                System.err.println( "a possible bad parent is: " +
//                                    possibleParent.getNodeId() );
                if ( possibleParent != node &&
                     possibleParent.receivesStripe( stripeToJoin ) &&
                     possibleParent.allowChild( node ) ) {

                    possibleParent.getTreeInfo( stripeToJoin ).addChild( node );
                    node.makeTreeInfo( stripeToJoin ).setParent( possibleParent );
                    joined = true;
//                    System.err.println( "Node found a BAD parent! Parent is: " +
//                                       possibleParent.getNodeId() );
                    break;
                } else {
//                    System.err.println( "possible BAD parent won't work, problem " +
//                                        "is:" );
//                    if ( possibleParent == node )
//                        System.err.println( "possible parent IS node" );
//                    if ( !possibleParent.receivesStripe( stripeToJoin ) )
//                       System.err.println( "possible parent doesn't get stripe" );
//                    if ( !possibleParent.allowChild( node ) )
//                        System.err.println( "poss. parent won't allow child" );
                }
            }
        }

        // if node still doesn't receive stripe it
        // didn't actually JOIN a stripe
        if ( !joined ) {
//            System.err.println( "Node couldn't find parent, so findParent is " +
//                                "about to return false....." );
            return false;
        } else {
//            System.err.println( "Node found a parent, so findParent will " +
//                                "return true and add it to inNetwork if it's " +
//                                "not already in there....." );
            if ( !inNetwork.contains( node ) )
                inNetwork.add( node );
            return true;
        }
    }

    /**
     * Returns the next node ID using the node ID allocation algorithm from this TreeManager.
     */
    public int getNextNodeId() {
        return nodeIdCounter++;
    }
}
