package Fairness.Trees;

import Fairness.Nodes.Node;
import Fairness.Simulation;
import org.gillius.Util.CheckUtils;

import java.util.Collection;
import java.util.HashSet;
import java.util.Set;
import java.util.Iterator;
import java.util.List;
import java.util.LinkedList;
import java.util.TreeSet;

/**
 * The TreeManager manages the construction of the stripe trees, and is also responsible
 * for assigning node IDs and stripe IDs.
 *
 * @author Jason Winnebeck
 */
public abstract class TreeManager {
    private Set<Node> nodes = new HashSet<Node>();
    protected Simulation sim;

    /**
     * Returns all nodes in the network.
     */
    protected Collection<Node> getAllNodes() {
        return nodes;
    }

    /**
     * Sets up the tree manager to run within the given simulation.
     *
     * @param sim non-null Simulation
     */
    public void setUp( Simulation sim ) {
        CheckUtils.checkNull( sim, "sim" );

        this.sim = sim;

        setUp();
    }

    /**
     * Called when the public setup method is called. At this time, the sim parameter has been
     * initialized, and getAllNodes also should be valid to call.
     */
    protected void setUp() {
    }

    /**
     * After a tree reconstruction, returns the nodes that are the roots of their stripe trees. It
     * is assumed that the Node is the root of its primary stripe and of no other. That is, Node
     * getParent( getStripeId() ) should return null.
     *
     * @return non-null Collection of 0 or more Node.
     */
    public abstract Collection<Node> getRootNodes();

    /**
     * A method to completely rebuild the Node tree given all nodes in the system.
     */
    public abstract void rebuildTree();

    /**
     * Adds a Node to this TreeManager for it to manage and build into trees.
     *
     * @param node non-null Node that has an ID that came from getNextNodeId.
     */
    public void addNode( Node node ) {
        CheckUtils.checkNull( node, "node" );

        nodes.add( node );
    }

    /**
     * Returns the next node ID using the node ID allocation algorithm from this TreeManager.
     *
     */
    public abstract int getNextNodeId();

    /**
     * Return a String version of the whole forest.
     */
    public String forestToString() {
        String retVal = "";

        Set<Integer> stripeIds = new TreeSet<Integer>();
        for ( Node node : getAllNodes() ) {
            stripeIds.addAll( node.getStripes() );
        }

        Iterator<Integer> stripeIter = stripeIds.iterator();

        // print out String block for each stripe tree in forest
        while ( stripeIter.hasNext() ) {

            int stripeId = stripeIter.next();
            Node theRoot = null;
            retVal += "Tree for stripeId: " + Integer.toHexString( stripeId );
            retVal += "\n";
            Iterator<Node> rootsIter = getRootNodes().iterator();
            while ( rootsIter.hasNext() ) {
                Node rootNode = rootsIter.next();
                if ( rootNode.receivesStripe( stripeId ) &&
                     rootNode.getTreeInfo( stripeId ).getParent() == null )
                    theRoot = rootNode;

            }

            retVal += "Root: " + theRoot.toString() + "\n";

            // after printing out root, we do a BFS of the stripe tree,
            // printing each level of the tree on a separate line;
            // on each level, the Strings of Nodes on that level are
            // printed, but before each such String is the String of that
            // Node's parent in ()'s
            List<Node> q = new LinkedList<Node>();
            Iterator<Node> childIter =
                theRoot.getTreeInfo( stripeId ).getChildren().iterator();
            while ( childIter.hasNext() ) q.add( childIter.next() );
            int curDepth = 1;
            while ( !q.isEmpty() ) {

                Node curNode = q.remove( 0 );
                int newDepth = curNode.getPathToRoot( stripeId ).size();
                if ( newDepth == curDepth )
                    retVal += "(" + curNode.getParent( stripeId ).toString() +
                              ") " + curNode.toString() + " ";
                else if ( newDepth > curDepth ) {
                    retVal += "\n(" +
                              curNode.getParent( stripeId ).toString() +
                              ") " + curNode.toString() + " ";
                    curDepth = newDepth;
                }
                Iterator<Node> aChildIter =
                    curNode.getTreeInfo( stripeId ).getChildren().iterator();
                while ( aChildIter.hasNext() ) q.add( aChildIter.next() );

            } // while q isn't empty

            // put a blank line between each String block for a stripe tree
            retVal += "\n\n";

        }

        return retVal;

    }
}
